@@ -11,7 +11,6 @@
  * REPRESENTATION OR WARRANTY OF ANY KIND CONCERNING THE MERCHANTABILITY
  * OF THIS SOFTWARE OR ITS FITNESS FOR ANY PARTICULAR PURPOSE.
  */
-#include <stdarg.h>
 #include <string.h>
 #include "utf.h"
 #include "utfdef.h"
@@ -25,7 +24,7 @@ utfrune(const char *s, Rune c)
 	int n;
 
 	if(c < Runesync)		/* not part of utf sequence */
-		return strchr(s, c);
+		return strchr(s, (char)c);
 
 	for(;;) {
 		c1 = *(uchar*)s;
