@@ -6,5 +6,6 @@ enum {
 	thechar = '6',
 	BigEndian = 0,
 	CacheLineSize = 64,
-	appendCrossover = 16
+	RuntimeGogoBytes = 64,
+	PCQuantum = 1
 };
