@@ -2,7 +2,7 @@
 // Use of this source code is governed by a BSD-style
 // license that can be found in the LICENSE file.
 
-// +build darwin freebsd linux netbsd openbsd windows
+// +build darwin dragonfly freebsd linux netbsd openbsd windows
 
 #include "runtime.h"
 
@@ -11,7 +11,8 @@ Slice syscall·envs;
 byte*
 runtime·getenv(int8 *s)
 {
-	int32 i, j, len;
+	int32 i, j;
+	intgo len;
 	byte *v, *bs;
 	String* envv;
 	int32 envc;
