@@ -4,7 +4,9 @@
 
 #include "runtime.h"
 #include "os_GOOS.h"
+#include "../../cmd/ld/textflag.h"
 
+#pragma textflag NOSPLIT
 int64
 runtime·nanotime(void)
 {
