@@ -1,4 +1,4 @@
-// +build amd64
+// +build 386 arm
 // errorcheck
 
 // Copyright 2011 The Go Authors.  All rights reserved.
@@ -6,17 +6,10 @@
 // license that can be found in the LICENSE file.
 
 // Issue 2444
-// Issue 4666: issue with arrays of exactly 4GB.
 
 package main
-
-func main() { // ERROR "stack frame too large"
-	var arr [1000200030]int32
+func main() {
+	var arr [1000200030]int   // ERROR "type .* too large"
 	arr_bkup := arr
 	_ = arr_bkup
 }
-
-func F() { // ERROR "stack frame too large"
-	var arr [1 << 30]int32
-	_ = arr[42]
-}
